#lang scheme

(require "structures.scm")
(require "parser.scm")
(require "concurrency.scm")
(require "util.scm")

(provide run-file run-files run)

(define root-jp (make-jp #f #f #f '() #f))

; evaluation

(define (run-file path)
  (eval-program (parse-file path)))

(define (run-files paths)
  (eval-program (parse-files paths)))

(define (run text)
  (eval-program (parse-program text)))

(define (eval-program decls)
  (empty-classes)
  (empty-threads)
  
  (init-classes! decls)
  (eval-expr
   (parse-expr '(send (new Main) main))
   (env-empty)
   '()
   root-jp)
  
  (wait-for-threads)
  (empty-classes)
  (empty-threads)
  0)

(define (eval-body body E A jp) (car (reverse (eval-exprs body E A jp)))) ; return last result
(define (eval-exprs exprs E A jp) (map (lambda(e)(eval-expr e E A jp)) exprs))

; todo: try to use 'match'
(define (eval-expr expr E A jp)
  ;(printf "eval: ~a~n" expr)
  ;(printf "eval: ~a~n" A)
  (cond
    ;lit and or cond get set if let prim-app
    ((lit-expr? expr)  (lit-expr-value expr))
    ((vget-expr? expr) (env-get  (vget-expr-name expr) E))
    ((vset-expr? expr) (env-set! (vset-expr-name expr) 
                                 (eval-expr (vset-expr-value expr) E A jp) 
                                 E))
    
    ((and-expr? expr)
     (let loop ((args (and-expr-args expr)))
       (if (null? args)
           #t
           (and (eval-expr (car args) E A jp)
                (loop (cdr args))))))
    
    ((or-expr? expr)
     (let loop ((args (or-expr-args expr)))
       (if (null? args)
           #f
           (or (eval-expr (car args) E A jp)
               (loop (cdr args))))))
    
    ((cond-expr? expr) (let loop ((branches (cond-expr-branches expr)))
                         (if (eval-expr (caar branches) E A jp)
                             (eval-expr (cadar branches) E A jp)
                             (if (null? (cdr branches))
                                 'undefined
                                 (loop (cdr branches))))))
    
    ((let-expr? expr) (eval-body (let-expr-body expr) 
                                 (env-extend (let-expr-names expr)
                                             (eval-exprs (let-expr-values expr) E A jp)
                                             E)
                                  A jp))
    
    ((while-expr? expr) (let loop ((rvalue 'undefined))
                          (if (eval-expr (while-expr-cond expr) E A jp)
                              (loop (eval-body (while-expr-body expr) E A jp))
                              rvalue)))
    
    ((prim-app-expr? expr) (apply (prim-app-expr-fun expr) 
                                  (eval-exprs (prim-app-expr-args expr) E A jp)))
    
    ; oo
    ; new get set send super
    ((new-expr? expr)(let* ((class (lookup-class (new-expr-class expr)))
                            (args  (eval-exprs (new-expr-args expr) E A jp))
                            (njp   (make-jp 'new #f class args jp)))
                       (weave-some A njp)
                       (let ((obj  (make-rt-object class 
                                                   (make-vector (length (accumulate-fields class)) 'undefined) 
                                                   (new-monitor)
                                                   (collect-match-d njp A))))
                         (exec-method (lookup-method 'init class) obj args A njp)
                         obj)))
    
    ((fget-expr? expr) (field-get (env-get 'this E)
                                  (fget-expr-name expr)
                                  (env-get '%host E)))
    
    ((fset-expr? expr) (field-set! (env-get 'this E)
                                   (fset-expr-name expr)
                                   (eval-expr (fset-expr-value expr) E A jp)
                                   (env-get '%host E)))
    
    ((send-expr? expr) (let* ((obj    (eval-expr (send-expr-target expr) E A jp))
                              (method (lookup-method (send-expr-name expr) (rt-object-class obj)))
                              (args   (eval-exprs (send-expr-args expr) E A jp)))
                         (let ((njp (make-jp 'call (send-expr-name expr) obj args jp)))
                           (weave-some A njp)
                           (exec-method method obj args (collect-match-c njp A) njp))))
    
    ((super-expr? expr) (let* ((obj    (env-get 'this E))
                               (method (lookup-method (super-expr-name expr) (rt-class-super (env-get '%host E))))
                               (args   (eval-exprs (super-expr-args expr) E A jp)))
                          (exec-method method obj args A jp)))
    
    ; concurrency
    ((thread-expr? expr) (let* ((obj    (eval-expr (thread-expr-target expr) E A jp))
                                (method (lookup-method 'run (rt-object-class obj))))
                           (augment-threads (thread (lambda()
                                                      (exec-method method obj '() A jp))))))
    
    ((wait-expr? expr) (monitor-wait (env-get 'this E) (wait-expr-cond-name expr)))
    
    ((notify-expr? expr) (monitor-notify (env-get 'this E) (notify-expr-cond-name expr)))
    
    ; aspects
    ((deploy-expr? expr) (eval-body (deploy-expr-body expr) E (new-dasps expr E A jp) jp))
    
    ((deploy-on-expr? expr) (let ((obj (eval-body (deploy-on-expr-body expr) E A jp)))
                              ;(printf "1: ~a~n" (rt-object-aspects obj))
                              (set-rt-object-aspects! obj (union (new-dasps expr E '() jp)
                                                                 (rt-object-aspects obj)))
                              ;(printf "2: ~a~n" (rt-object-aspects obj))
                              obj))
    
    (else
     (error 'eval-expr "~a~n" expr))
    ))	

(define (exec-method method obj args A jp)
  (if (not (= (length args) (length (rt-method-params method))))
      (error 'exec-method "wrong number of arguments: (~a.~a) ~a =/= ~a" 
             (rt-class-name (rt-method-class method)) (rt-method-name method) (rt-method-params method) args)
      (begin
        (let ((njp   (make-jp 'exec (rt-method-name method) obj args jp)))
          (weave-some (union A (rt-object-aspects obj)) njp)
          ; sync
          (cond ((rt-method-sync method) (monitor-enter obj))) 
          
          (let ((result (eval-body (rt-method-body method)
                                   (env-extend (append (list 'this '%host) (rt-method-params method))
                                               (append (list  obj   (rt-method-class method)) args) 
                                               (env-empty))
                                   (union A (rt-object-aspects obj)) ; recalculate: an aspect can augment the obj.aspects set
                                   njp)))
            
            ; sync
            (cond ((rt-method-sync method) (monitor-leave obj)))
            result)))))

; aspects 

(define (new-dasps expr E A jp)
  (append (map (lambda(e)(new-dasp e E jp)) 
               (if (deploy-expr? expr)
                   (deploy-expr-dyn-aspects expr)
                   (deploy-on-expr-dyn-aspects expr))) 
          A))

(define (new-dasp expr E jp)
  (make-dasp (dyn-aspect-c expr)
             (dyn-aspect-d expr)
             (dyn-aspect-f expr)
             (eval-expr (dyn-aspect-aspect expr) E '() jp)))

(define (weave-some asps jp)
  (weave-all (collect-match-f jp asps) jp))

(define (weave-all asps jp)
  (map (lambda (x) (weave x jp)) (map dasp-aspect asps)))

(define (weave asp jp)
;  (cond (not (lookup-method-no-error mname (rt-object-class asp)))
;        (printf 'weave "method not found: ~a~n" mname)
;        #f)
  ; if (asp.pc(jp) { asp.advice(jp) }
  (let ((pc-method  (lookup-method 'pc (rt-object-class asp)))
        (adv-method (lookup-method 'advice (rt-object-class asp))))
    (cond ((eval-body (rt-method-body pc-method) 
                      (env-extend (list 'this '%host 'jp) (list asp (rt-method-class pc-method) jp) (env-empty)) '() jp)
           (eval-body (rt-method-body adv-method) 
                      (env-extend (list 'this '%host 'jp) (list asp (rt-method-class adv-method) jp) (env-empty)) '() jp)))))

(define (collect-match-c jp asps) (filter (lambda (a) (app/prim (dasp-c a) jp)) asps))
(define (collect-match-d jp asps) (filter (lambda (a) (app/prim (dasp-d a) jp)) asps))
(define (collect-match-f jp asps) (filter (lambda (a) (app/prim (dasp-f a) jp)) asps))

;; applies the (pointcut) p, passing jp as argument
;; evaluation is performed in an empty aspect environment
(define (app/prim p jp)
  (eval-expr p (env-extend (list 'jp) (list jp) (env-empty)) '() jp))